#!/usr/bin/python
# Copyright (c) 2004-2019 Bright Computing Holding BV. All Rights Reserved.
#
# This software is the confidential and proprietary information of
# Bright Computing Holding BV ("Confidential Information").  You shall not
# disclose such Confidential Information and shall use it only in
# accordance with the terms of the license agreement you entered into
# with Bright Computing Holding BV or its subsidiaries.


import sys
import os
import re
import threading
import logging
import time
import shutil
import socket
from stat import ST_MODE, S_ISBLK

from subprocess import Popen, PIPE

progresschars = ['[ .    ]', '[ ..   ]', '[ ...  ]', '[ .... ]']


def run_thread(work):
    ts = work()
    ts.start()
    alive = ts.isAlive()
    prid = 0
    while alive:
        alive = ts.isAlive()
        prid = print_progress(prid)
    return ts.join()


def print_progress(pos):
    sys.stdout.write(progresschars[pos])
    time.sleep(.25)
    sys.stdout.write('\b\b\b\b\b\b\b\b')
    sys.stdout.flush()
    pos += 1
    pos %= len(progresschars)
    return pos


def print_header(msg):
    m = msg.ljust(45 + 1, '.') + "... "
    sys.stdout.write(m)
    # print msg.ljust(45 + 1, '.') + "... ",
    sys.stdout.flush()
    logger.Log(msg)


def print_result(result):
    if result:
        print("[  OK  ]")
    else:
        print("[FAILED]")


def invoke_call(header, thr):
    result = True
    print_header(header)
    result = run_thread(thr)
    print_result(result)
    return result


def cleanup_and_exit(gd):
    cleanup_dvd_repo(gd)
    if len(gd.errors) > 0:
        print_error('', gd, exit=True)
    sys.exit(1)


def tee(msg, color=None):
    try:
        if color is None:
            print('%s %s %s' % (Font.BOLD, msg, Font.CLOSE))
        else:
            print('%s %s %s' % (color, msg, Font.CLOSE))
        logger.Log(msg)
    except Exception as e:
        logger.LogE(str(e), True)


def un_mount(p, gd):
    good = True
    try:
        rc = RunCommand()
        rc.run(['/bin/umount', p])
        good = rc.good
    except Exception as e:
        gd.errors.append(str(e))
    return good


def stop_service(proglist, service):
    res = True
    try:
        # Do not rely on status return code. It is    unreliable sometimes and
        # also the option does not exist for certain scripts.
        if service_available(service):
            rc = RunCommand()
            if os.path.exists('/sbin/service'):
                rc.run(['/sbin/service', service, 'stop'])
            else:
                rc.run(['systemctl', 'stop', service])
            if not rc.good:
                logger.Log('Command failed: service %s stop' % (service))
            r = wait_for_service_stop(proglist)
            if not r:
                res = False
        else:
            logger.Log('Will not attempt to stop service %s (Not available)' % (service))
    except Exception as e:
        logger.LogE(str(e))
        res = False
    return res


def start_service(service, wait=None, start='start'):
    res = True
    try:
        rc = RunCommand()
        if os.path.exists('/usr/lib/systemd/system'):
            rc.run(['/usr/bin/systemctl', start, service])
        else:
            rc.run(['/sbin/service', service, start])
        res = rc.good
        if wait is not None:
            logger.Log('Waiting for %s to complete init' % (service))
            time.sleep(wait)
    except Exception as e:
        logger.LogE(str(e))
        res = False
    return res


def service_available(service):
    res = False
    try:

        return (os.path.exists('/etc/init.d/%s' % (service))
                or os.path.exists('/usr/lib/systemd/system/%s.service' % (service))
                or os.path.exists('/lib/systemd/system/%s.service' % (service)))

    except Exception as e:
        logger.LogE(str(e))
        res = False
    return res


def service_running(service):
    r = False
    try:
        rc = RunCommand()
        if os.path.exists('/usr/lib/systemd/system'):
            rc.run(['/usr/bin/systemctl', 'status', service])
        else:
            rc.run(['/sbin/service', service, 'status'])

        return rc.good

    except Exception as e:
        logger.Log('Unable to get service status for %s : %s' % (service, str(e)))
        r = False
    return r


def disable_service(service):
    r = False
    try:
        rc = RunCommand()
        rc.run(['/usr/bin/systemctl', 'disable', service])
        return rc.good

    except Exception as e:
        logger.Log('Unable to disable service %s: %s' % (service, str(e)))
        r = False
    return r


def mask_service(service, options=None):
    r = False
    try:
        rc = RunCommand()
        if options is None:
            rc.run(['/usr/bin/systemctl', 'mask', service])
        else:
            rc.run(['/usr/bin/systemctl', 'mask', options, service])
        return rc.good

    except Exception as e:
        logger.Log('Failed to mask service %s: %s' % (service, str(e)))
        r = False
    return r


def unmask_service(service, options=None):
    r = False
    try:
        rc = RunCommand()
        if options is None:
            rc.run(['/usr/bin/systemctl', 'unmask', service])
        else:
            rc.run(['/usr/bin/systemctl', 'unmask', options, service])
        return rc.good

    except Exception as e:
        logger.Log('Failed to unmask service %s: %s' % (service, str(e)))
        r = False
    return r


def process_running(process):
    r = False
    try:
        pidof = '/sbin/pidof' if os.path.exists('/sbin/pidof') else '/bin/pidof'
        if os.system("%s %s 1>/dev/null" % (pidof, process)) == 0:
            r = True
    except Exception as e:
        logger.LogE(str(e))
        r = False
    return r


def wait_for_service_stop(proglist):
    r = True
    try:
        retries = 0
        for prog in proglist:
            if process_running(prog):
                logger.Log('Checking: ' + prog)
                alive = 0
                while alive == 0 and retries < 5:
                    alive = os.system('/sbin/pidof ' + prog + " 1>/dev/null")
                    retries += 1
                    time.sleep(2)
                if alive == 0:
                    logger.Log('Process still running: ' + prog)
                    r = False
    except Exception as e:
        logger.LogE(str(e))
        r = False
    return r


def copy_client_certs():
    try:
        client_pem_new = '/root/.cm/admin.pem'
        client_key_new = '/root/.cm/admin.key'
        client_pem_old = '/root/.cm/cmsh/admin.pem'
        client_key_old = '/root/.cm/cmsh/admin.key'
        cmsh_history_old = '/root/.cm/cmsh/.cmshhistory'
        cmsh_history_new = '/root/.cm/.cmshhistory'
        if not os.path.isfile(client_pem_new):
            if os.path.isfile(client_pem_old):
                shutil.copy(client_pem_old, client_pem_new)
                os.system('chmod 600 ' + client_pem_new)
        if not os.path.isfile(client_key_new):
            if os.path.isfile(client_key_old):
                shutil.copy(client_key_old, client_key_new)
                os.system('chmod 600 ' + client_key_new)
        if not os.path.isfile(cmsh_history_new):
            if os.path.isfile(cmsh_history_old):
                shutil.copy(cmsh_history_old, cmsh_history_new)
                os.system('chmod 600 ' + cmsh_history_new)

    except Exception as e:
        logger.LogE(str(e))


def skip_line_cmdaemon_conf(line):
    r = [None, None]
    try:
        if len(line) <= 1 or line.startswith('#') or empty_string(line):
            return r
        elif line.startswith('Service {') or line.startswith('}'):
            return r
        elif '=' in line:
            (var, value) = line.split('=', 1)
            var = var.strip(' ')
            value = value.strip(' ')
            if var == 'Path' or var == 'Public' or var == 'ObjectFile':
                return r
            return [var, value]
    except Exception as e:
        logger.LogE(str(e))
        return r
    return r


def skip_line_node_installer_conf(line):
    r = [None, None]
    try:
        if len(line) <= 1 or line.startswith('#') or empty_string(line):
            return r
        elif '=' in line:
            (var, value) = line.split('=', 1)
            var = var.strip(' ')
            value = value.strip(' ')
            return [var, value]
    except Exception as e:
        logger.LogE(str(e))
        return r
    return r


def read_node_installer_conf(conf_file):
    conf = {}
    try:
        f = open(conf_file, 'r')
        for line in f:
            line = line.strip('\n')
            (var, value) = skip_line_node_installer_conf(line)
            if var is not None and value is not None:
                conf[var] = value
        f.close()
        return conf
    except Exception as e:
        logger.LogE(str(e))
        conf = {}
    return conf


def read_cmdaemon_conf(conf_file):
    conf = {}
    try:
        f = open(conf_file, 'r')
        for line in f:
            line = line.strip('\n')
            (var, value) = skip_line_cmdaemon_conf(line)
            if var is not None and value is not None:
                conf[var] = value
        f.close()
        return conf
    except Exception as e:
        logger.LogE(str(e))
        conf = {}
    return conf


def read_cmd_advanced_config(cmd_conf_file, adv_config_flag):
    try:
        cmdaemon_conf = read_cmdaemon_conf(cmd_conf_file)

        if 'AdvancedConfig' not in cmdaemon_conf:
            return None

        advanced_config = cmdaemon_conf['AdvancedConfig'].replace('{', '').replace('}', '')
        for param in advanced_config.split(','):
            if adv_config_flag in param:
                return param.split('=')[1].strip('"')
        return None

    except Exception as e:
        logger.LogE(str(e))


def skip_line_generic_conf(line):
    r = [None, None]
    try:
        if len(line) <= 1 or line.startswith('#') or empty_string(line):
            return r
        elif '=' in line:
            (var, value) = line.split('=', 1)
            var = var.strip(' ')
            value = value.strip(' ')
            return [var, value]
    except Exception as e:
        logger.LogE(str(e))
        return r
    return r


def read_conf_generic(conf_file):
    conf = {}
    try:
        f = open(conf_file, 'r')
        for line in f:
            line = line.strip('\n')
            (var, value) = skip_line_generic_conf(line)
            if var is not None and value is not None:
                conf[var] = value
        f.close()
        return conf
    except Exception as e:
        logger.LogE(str(e))
        conf = {}
    return conf


def cleanup_dvd_repo(gd, d=None):
    try:
        if gd.use_iso_repo:
            if d is not None:
                """ Clean up bind mount in software images """
                if not un_mount('%s/%s' % (d, gd.iso_repo_dir), gd):
                    gd.errors.append('Cleanup: unable to unmount %s/%s' % (d, gd.iso_repo_dir))
                if os.path.exists(d + '/' + gd.dvd_repo_file):
                    os.remove(d + '/' + gd.dvd_repo_file)
            else:
                if not gd.do_not_unmount and d is None:
                    logger.Log('Will unmount DVD %s' % (gd.iso_repo_dir))
                    if not un_mount(gd.iso_repo_dir, gd):
                        gd.errors.append('Cleanup: unable to unmount temp directory %s' % (gd.iso_repo_dir))
                    else:
                        shutil.rmtree(gd.iso_repo_dir)
                if os.path.exists(gd.dvd_repo_file):
                    os.remove(gd.dvd_repo_file)
    except Exception as e:
        logger.LogE(str(e))


def print_error(msg, gd, exit=True):
    gd.errors = [x for x in gd.errors if x != '']
    errors_as_string = '\n'.join(gd.errors)
    logger.Log(msg)
    header = '%s ERROR(S): %s' % (Font.BOLD, Font.CLOSE)
    full_error = """
%s
%s
%s
%s
%s
    """ % (header, Font.RED, msg, errors_as_string, Font.CLOSE)
    print(full_error)

    if exit:
        sys.exit(1)


def print_warning(msg, exit=False):
    logger.Log(msg)
    header = '%s WARNINGS(S): %s' % (Font.BOLD, Font.CLOSE)
    warning = """
%s
%s
%s
%s
    """ % (header, Font.RED, msg, Font.CLOSE)
    print(warning)

    if exit:
        sys.exit(1)


def print_exception(msg, exit=True):
    try:
        exc_type, exc_obj, exc_tb = sys.exc_info()
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
        print("""
EXCEPTION(%s: %s):
%s
        """ % (fname, str(exc_tb.tb_lineno), msg))
    except Exception as e:
        print("""
EXCEPTION:
%s
        """ % (msg + '\n' + str(e)))
    if exit:
        sys.exit(1)


def empty_string(mystr):
    return (len(mystr.strip(' ')) == 0)


def secure_copy(src, dst):
    try:
        rc = RunCommand()
        _command = 'scp %s %s' % (src, dst)
        rc.run(_command.split())
        return rc.good

    except Exception as ex:
        logger.LogE(str(ex))
        return False


def is_reachable(host=None, port=None):
    hostip = None
    # Check if hostname is resolvable
    try:
        hostip = socket.gethostbyname(host)
    except Exception as ex:
        logger.LogE(str(ex))
        return False

    if (hostip is not None) and (port is not None):
        # Check if host is reachable
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(5)
            sock.connect((hostip, int(port)))
            sock.close()
            return True
        except Exception as ex:
            logger.LogE(str(ex))
            return False


def is_block_device(dev, gd):
    try:
        if os.path.exists(dev):
            mode = os.stat(dev)[ST_MODE]
            if S_ISBLK(mode):
                return True
            else:
                logger.Log(dev + " is not a block special device file.")
                return False
        else:
            log_and_append("Cannot find path " + dev, gd.errors)
            return False
    except Exception as e:
        logger.LogE(str(e))
        return False


def log_and_append(msg, errors):
    errors.append(msg)
    logger.Log(msg)


class Font:
    PURPLE = '\033[95m'
    CYAN = '\033[96m'
    DARKCYAN = '\033[36m'
    BLUE = '\033[94m'
    GREEN = '\033[32m'
    YELLOW = '\033[93m'
    RED = '\033[91m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    CLOSE = '\033[0m'


class Logger:
    def __init__(self, logfile):
        self.logfile = logfile
        self.logger = logging.getLogger('cmupgradelog')
        self.hdlr = logging.FileHandler(logfile)
        self.formatter = logging.Formatter('(%(asctime)s) %(message)s')
        self.hdlr.setFormatter(self.formatter)
        self.logger.addHandler(self.hdlr)
        self.logger.setLevel(logging.INFO)

    def Log(self, msg, lspace=0):
        for x in msg.split('\n'):
            if not empty_string(x):
                self.logger.info(x)

    def LogV(self, msg, gd, lspace=0):
        if gd.verbose:
            for x in msg.split('\n'):
                if not empty_string(x):
                    self.logger.info('VERBOSE: ' + x)

    def LogE(self, msg, lspace=0):
        exc_type, exc_obj, exc_tb = sys.exc_info()
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
        self.Log('Exception in ' + fname + '(line no:' + str(exc_tb.tb_lineno) + '): ' + msg)


def replace_pattern_in_file(file_name, old_pattern, new_pattern):
    try:
        input_file = open(file_name, 'r')
        output_file = open(file_name + '.new', 'wt')
        for line in input_file:
            line = line.strip('\n')
            pattern_found = re.search(old_pattern, line)
            if pattern_found:
                line = re.sub(r"%s" % (old_pattern), new_pattern, line)
            output_file.write(line + '\n')
        output_file.close()
        input_file.close()
        shutil.move('%s.new' % (file_name), file_name)

    except Exception as e:
        logger.Log(str(e))


def read_file(mfile):
    data = ''
    try:
        f = open(mfile, 'r')
        data = f.read()
        return data
    except Exception as e:
        logger.Log(str(e))
    return data


def current_package_version(package_name, target):
    version = None
    try:
        current_link = '/cm/%s/apps/%s/current' % (target, package_name)
        if os.path.islink(current_link):
            _version = os.path.basename(os.readlink(current_link).split('_')[0])
            return '.'.join(re.sub("([a-zA-Z]+)", "", _version).split('.')[0:2])

    except Exception as e:
        logger.Log(str(e))
    return version


def _lock_slurm_upgrades():
    try:

        slurm_version = current_package_version('slurm', 'shared')
        if slurm_version is None:
            logger.Log("No slurm version found, will not lock from upgrades")
            return False

        if float(slurm_version) <= 16.05:
            logger.Log("Slurm version is %s, not eligible for upgrade" % (str(slurm_version)))
            return True
        logger.Log("Slurm version is %s, eligible for upgrade" % (str(slurm_version)))
        return False

    except Exception as e:
        logger.LogE(str(e))
        return False


def connect_to_mysql(gd):
    try:
        cmdaemon_conf = read_cmdaemon_conf(gd.cmd_conf)

        mysql = CmMySQLDb(gd, host=cmdaemon_conf['DBHost'].strip('"'),
                          user=cmdaemon_conf['DBUser'].strip('"'),
                          passwd=cmdaemon_conf['DBPass'].strip('"'),
                          database=cmdaemon_conf['DBName'].strip('"'))

        if mysql is None:
            logger.Log('unable to connect to mysql')
        elif mysql.connection is None:
            logger.Log('\n'.join(mysql.errors))

    except Exception as e:
        logger.LogE(str(e))
        logger.Log('\n'.join(mysql.errors))

    return mysql


def preliminary_cleanup():
    # This file is known to be useless and to cause service pbs status
    # to fail in our installations
    # nevertheless, it is copied there during pbspro setup, therefore
    # it must be removed here
    pbs = '/etc/init.d/pbs'
    if os.path.isfile(pbs):
        os.remove(pbs)


def fix_current_symlink(package_name, package_base_dir):
    try:
        logger.Log('Checking current symlink for package {}'.format(package_name))
        rc = RunCommand()
        package_version = rc.check_output(['rpm', '-q', '--qf', '%{VERSION}', package_name])
        if package_version is None:
            logger.Log('Cannot determine version for package {}'.format(package_name))
            return
        package_current_link = '{}/{}'.format(package_base_dir, 'current')
        if os.path.islink(package_current_link):
            if package_version != os.readlink(package_current_link):
                os.unlink(package_current_link)
        if not os.path.exists(package_current_link):
            logger.Log('Fixing current symlink for package {}:{}'.format(package_name, package_version))
            if os.path.isdir('{}/{}'.format(package_base_dir, package_version)):
                os.symlink(package_version, '{}/{}'.format(package_base_dir, 'current'))

    except Exception as e:
        logger.LogE(str(e))


class Help:
    post_upgrade = """\
Re-run only post upgrade actions that involve updating the
cmdaemon configuration. This should be used only when the
upgrade itself has completed succesfully, but some settings
like the workload manager and cloud node settings could not
be updated during the regular run, due to cmdaemon not being
reachable
    """
    upgrade_head_node = """\
Upgrade Bright head node
    """
    upgrade_software_images = """\
List of software images names (comma separated)
    """
    bright_dvd = """\
Path to the Bright ISO/DVD to be used for upgrade
    """
    disable_repos = """\
Disable repositories during upgrade. This is helpful in
preventing yum from failing when some of the configured
repositories are not reachable (comma seperated list of repo
ids or glob or a combination of both)
    """
    image_base_dirs = """\
List of additional directories to find software images if
they are in not in the standard location: /cm/images
(comma separated)
    """
    skip_cmshared_check = """\
Skip checking if /cm/shared is mounted. Use this option to
do in-place upgrade when /cm/shared is mounted from an
external source in a high-availability configuration
    """
    ignore_openstack_check = """\
Skip checking if OpenStack is installed
    """
    skip_cloud_check = """\
Skip checking if a cloud setup has been deployed
    """
    skip_hadoop_setup_check = """\
Skip checking if Hadoop has been deployed
    """
    do_not_stop_wlm = """
Do not stop workload manager services. When this option is
used, the --skipnodestatescheck is enabled automatically.
This allows for compute nodes to be UP during upgrade with
no interruption to running workload manager jobs
    """
    verbose = """\
Print verbose output
    """
    non_interactive = """\
Run in non interactive mode, no questions will be asked
    """
    cuda_upgrade = """\
Also upgrade CUDA packages
    """
    force_upgrade = """\
Force re-run of upgrade. Useful if the upgrade had been
interrupted or had failed unexpectedly
    """
    ignore_wlm_running_check = """\
Allow upgrade to proceed even if some workload management
processes could not be stopped
    """
    skip_node_states_check = """\
Skip checking nodes states, this is useful when cmdaemon
is in an unreachable state
    """
    skip_kubernetes_check = """\
Skip checking if kubernetes deployment exists
    """
    update_node_installer = """\
Update node-installer image
    """
    no_prereq_repos = """\
Use this option to prevent cm-upgrade from configuring
external prerequisite repositories
    """


class RunCommand:
    def __init__(self):
        self.command = []
        self.good = True
        self.proc = None

    def run(self, cmd, shell=False):
        self.command = cmd
        log = open(logger.logfile, "a")
        # log.write('Run command: ' + ' '.join(self.command) + '\n')
        os.environ['DEBIAN_FRONTEND'] = 'noninteractive'
        os.environ.update()
        self.proc = Popen(self.command, stdout=log, stderr=log, env=dict(os.environ), shell=shell)
        self.proc.communicate()
        if self.proc.returncode != 0:
            self.good = False
            # log.write('Command (' + ' '.join(self.command) + ') failed with exit code: ' + str(self.proc.returncode) + "\n")
            log.write('Command (%s) failed with exit code: %s\n' % (self.command, str(self.proc.returncode)))
        else:
            self.good = True
        log.close()

    def check_output(self, cmd):
        try:
            self.command = cmd
            self.proc = Popen(self.command, stdout=PIPE, stderr=PIPE, universal_newlines=True)
            output = self.proc.communicate()
            if self.proc.returncode != 0:
                return None
            return output[0]
        except Exception as e:
            logger.LogE(str(e))
            return None


class CmMySQLDb:

    def __init__(self, gd, host='localhost', user='', passwd='', database='', connect=True):

        self.cursor = None
        self.connection = None
        self.host = host
        self.user = user
        self.passwd = passwd
        self.database = database
        self.errors = []
        self.output = None
        self.query_success = None
        self.gd = gd

        if connect:
            self.__connect_to_db()

    def __connect_to_db(self):
        """ Only accessible inside class CmMySQLDb """
        try:
            if self.gd.is_rhel8():
                import pymysql as MySQLdb
            else:
                import MySQLdb
            self.connection = MySQLdb.connect(self.host,
                                              self.user,
                                              self.passwd,
                                              self.database)

            self.cursor = self.connection.cursor()

        except Exception as ex:
            self.errors.append(str(ex))

        """ Retry with TCP socket """
        if (self.connection is None) and (self.host in ('localhost', '')):
            try:
                if self.gd.is_rhel8():
                    import pymysql as MySQLdb
                else:
                    import MySQLdb
                self.connection = MySQLdb.connect('127.0.0.1',
                                                  self.user,
                                                  self.passwd,
                                                  self.database)

                self.cursor = self.connection.cursor()
                self.errors.pop()

            except Exception as ex:
                self.errors.append(str(ex))

    def disconnect(self):
        self.__disconnect_from_db()

    def __disconnect_from_db(self):
        """ Only accessible inside class CmMySQLDb """
        try:

            self.cursor.close()
            self.connection.close()

        except Exception as ex:
            self.errors.append(
                'Internal error while disconnecting from MySQL: ' + str(ex))

    def run_query(self, query, get_output=False):
        self.query_success = None

        try:
            if self.cursor.execute(query):
                self.query_success = True
                if get_output:
                    self.output = self.cursor.fetchall()
            else:
                self.query_success = False
                self.errors.append('Query failed: %s' % (query))

        except Exception as ex:
            self.errors.append(str(ex))


class MyThread(threading.Thread):
    def join(self):
        super(MyThread, self).join()
        return self.result


class GlobalDefs:
    def __init__(self):
        self.upgrade_version = '9.1'
        self.cmd_dir = '/cm/local/apps/cmd'
        self.cmd_bin = self.cmd_dir + '/sbin/cmd'
        self.cmd_etc_dir = self.cmd_dir + '/etc'
        self.cmd_conf = self.cmd_etc_dir + '/cmd.conf'
        self.cmupgrade_dir = '/cm/local/apps/cm-upgrade/%s' % (self.upgrade_version)
        self.cmupgrade_conf_dir = '%s/conf' % (self.cmupgrade_dir)
        self.default_monitoring = '%s/monitoring-default.xml' % (self.cmupgrade_conf_dir)
        self.nd_conf_file = '/cm/node-installer/scripts/node-installer.conf'
        self.cm_release_file = '/etc/cm-release'
        self.cmd_state_file = '/var/spool/cmd/state'
        self.softwareimages_maindir = '/cm/images'
        self.additional_image_base_dirs = []
        self.iso_rpms_dir = '/data/packages'
        self.upgrade_version_release = 'cm%s' % (self.upgrade_version)
        self.current_image_name = None
        self.current_image = None
        self.base_dist = None
        self.base_dist_generic = None
        self.update_images = False
        self.update_node_installer = False
        self.node_installer_dir_orig = '/cm/node-installer'
        self.node_installer_dir_new = '/cm/node-installer.new'
        self.cmdaemon_state = None
        self.images_to_update = []
        self.use_iso_repo = False
        self.use_prereq_repos = True
        self.iso_repo_path = None
        self.non_upgraded_packages = []
        self.valid_cmdaemon_states = ['ACTIVE', 'PASSIVE']
        self.package_info = None
        self.slurm_conf = '/cm/shared/apps/slurm/var/etc/slurm.conf'
        self.slurmdbd_conf = '/cm/shared/apps/slurm/var/etc/slurmdbd.conf'
        self.update_slurm_params = {"9.1": {"PrologSlurmctld": {"old": "/cm/local/apps/cmd/scripts/prolog-healthchecker,/cm/local/apps/cmd/scripts/prolog", "new": "/cm/local/apps/cmd/scripts/prolog-prejob"}}}
        self.update_slurm_pid_params = {
            self.slurm_conf: {
                "SlurmctldPidFile": {"/usr/lib/systemd/system/slurmctld.service": "PIDFile"},
                "SlurmdPidFile": {"/usr/lib/systemd/system/slurmd.service": "PIDFile"}
            },
            self.slurmdbd_conf: {
                "PidFile": {"/usr/lib/systemd/system/slurmdbd.service": "PIDFile"}
            }
        }
        self.update_nd_conf_params = {"9.1": {"privateKeyBits": {"old": "512", "new": "2048"}}}
        self.verbose = False
        self.non_interactive = False
        self.iso_repo_dir = None
        self.upgrade_excludes = 'pbspro*'
        self.upgrade_excludes_generic = ['mariadb*']
        self.stable_release_num = None
        self.minimum_stable_release = {"9.1": "9.1-7"}
        self.init_complete = False
        self.cm_config_cm = 'cm-config-cm'
        self.package_erase_nodeps = 'rpm -e --nodeps'.split(' ')
        self.package_erase_nodeps_noscripts = 'rpm -e --nodeps --noscripts'.split(' ')
        self.rpm_install = 'rpm -i'.split(' ')
        self.disable_repos_list = []
        self.post_upgrade = False
        self.ignore_openstack_check = False
        self.skip_cloud_setup_check = False
        self.skip_hadoop_setup_check = False
        self.do_not_stop_wlm = False
        self.do_not_unmount = False
        self.extra_package_groups = ['default', 'hpc', 'ofedstacks']
        self.prerequisites = ['python-boto']
        self.upgrade_backup_suffix = '.upgrade_backup'
        self.upgrade_new_suffix = '.upgrade_new'
        self.node_certificates_dir = '/cm/node-installer/certificates'
        self.errors = []
        self.sge_root = '/cm/shared/apps/sge/current'
        self.torque_root = '/cm/shared/apps/torque/current'
        self.pbspro_root = '/cm/shared/apps/pbspro/current'
        self.slurm_root = '/cm/shared/apps/slurm/current'
        self.wlm_services_disable = {'head_node': ['pbspro', 'pbs']}
        self.wlm_services = {"sgemaster.sge1": [self.sge_root + '/bin/linux-x64/sge_qmaster'],
                             "sgeexecd": [self.sge_root + '/bin/linux-x64/sge_execd'],
                             "moab": [],
                             "torque_server": [self.torque_root + '/sbin/pbs_server'],
                             "torque_mom": [self.torque_root + '/sbin/pbs_mom'],
                             "torque_sched": [self.torque_root + '/sbin/pbs_sched'],
                             "trqauthd": [self.torque_root + '/sbin/trqauthd'],
                             "pbs": [self.pbspro_root + '/sbin/pbs_server',
                                     self.pbspro_root + '/sbin/pbs_mom',
                                     self.pbspro_root + '/sbin/pbs_server.bin'],
                             "slurm": [self.slurm_root + '/sbin/slurmctld',
                                       self.slurm_root + '/sbin/slurmd'],
                             "slurmdbd": [self.slurm_root + '/sbin/slurmdbd']}
        self.wlm_services_extra = {"slurmctld": [self.slurm_root + '/sbin/slurmctld']}
        self.rpm_new_files = []
        self.exclude_from_updates = {'cmgui-ddn': ['cmgui']}
        self.valid_dist_path = {'sles12sp4': ['sles/12', 'sles/12sp4'],
                                'sles12sp3': ['sles/12sp3'],
                                'sles12sp5': ['sles/12sp5'],
                                'sles15sp1': ['sles/15sp1'],
                                'sles15sp2': ['sles/15sp2'],
                                'rhel7': ['rhel/7u6', 'rhel/7u7', 'rhel/7u8', 'rhel/7u9'],
                                'centos7': ['centos/7u6', 'centos/7u7', 'centos/7u8', 'centos/7u9'],
                                'rhel8': ['rhel/8u0', 'rhel/8u1', 'rhel/8u2', 'rhel/8u3'],
                                'centos8': ['centos/8u0', 'centos/8u1', 'centos/8u2', 'centos/8u3'],
                                'ubuntu1804': ['ubuntu/1804']}
        self.unsupported_dists = ['RHEL5', 'CENTOS5', 'SL5', 'SLES11', 'RHEL6', 'CENTOS6', 'SL6', 'SL7']
        self.cuda_upgrade = False
        self.wlm_upgrade = True
        self.force_upgrade = False
        self.cmd_start_wait = 30
        self.max_len = 80
        self.skip_node_states_check = False
        self.skip_cmshared_mounted_check = False
        self.ignore_wlm_running_check = False
        self.skipOldCmdConfDirectives = ['EnableJSON', 'EnableShellService']
        self.forceReplaceCmdConfDirectives = ['LDAPHost']
        self.cmd_backup_file = '/var/spool/cmd/cmd-backup-pre-upgrade.xml'
        self.skip_kubernetes_check = False
        # self.lock_slurm_upgrades = _lock_slurm_upgrades()
        self.lock_slurm_upgrades = False

    def initialize_progress_headers(self):
        self.AE_HEADER = 'Apply existing updates'
        self.BC_HEADER = 'Backup current configuration'
        self.UR_HEADER = 'Updating repository configuration'
        self.IP_HEADER = 'Install prerequisite packages'
        self.UI_HEADER = 'Setting up upgrade'
        self.UCM_HEADER = 'Update cm-config-cm'
        self.RD_HEADER = 'Removing pre-upgrade conflicts'
        self.CU_HEADER = 'Upgrading packages to Bright %s' % (self.upgrade_version)
        self.RO_HEADER = 'Removing obsoleted packages'
        self.UN_HEADER = 'Installing new required packages'
        self.RE_HEADER = 'Re-installing required Bright packages'
        self.CUDA_HEADER = 'Upgrading CUDA packages'
        self.RCU_HEADER = 'Disabling upgrade mode'
        self.PRN_HEADER = 'Processing config file changes'
        self.CREATE_RAMDISK = 'Re-creating ramdisk'
        self.RR_HEADER = 'Restoring backed-up files'
        self.MOTD_HEADER = 'Updating /etc/motd file'
        self.UPDATE_LDAP = 'Updating LDAP configuration'
        self.UPDATE_WLM_CONF = 'Updating Slurm configuration'
        self.UPDATE_WLM_AUTO = 'Updating workload manager autostart'
        self.GE_UPD = 'Updating Grid Engine configuration'
        self.PBS_UPD = 'Updating PBS Pro roles configuration'
        self.INITIALIZE_MONITORING = 'Initializing monitoring configuration'
        self.UPR_HEADER = 'Updating provisioning nodes'
        self.DOC_HEADER = 'Deleting old node certificates'
        self.REPOS_REACHABLE = 'Validating Bright %s repositories' % (self.upgrade_version)
        self.NODES_UP = 'Checking if nodes are still UP'
        self.STOP_SERV = 'Stopping cmd and workload manager services'
        self.SS_MNT = 'Checking if /cm/shared is mounted'
        self.MET_UPD = 'Updating metric collections'
        self.CHK_REQ = 'Checking required dist packages'
        self.UPDATE_NONUPG = 'Verifying upgraded packages'
        self.UPDATE_CLOUD = 'Updating cloud nodes settings'
        self.UPDATE_SWIMAGES = 'Updating software images'
        self.FETCH_CLUSTER_CERTS = 'Fetching cluster certificates'
        self.UPGRADE_POSSIBLE = 'Checking if upgrade is possible'
        self.UPD_MON = 'Updating monitoring configuration'
        self.NI_BACKUP_HEADER = 'Backing up node installer image'
        self.NI_UPGRADE_PREPARE = 'Preparing for upgrade'
        self.NI_UPGRADE = 'Installing new packages in node-installer'
        self.NI_ENABLE = 'Enabling new node-installer image'
        self.NI_POSTINSTALL = 'Running node-installer post install'
        self.FSPART_HEADER = 'Triggering fspart updates'

    def initialize_common_vars(self, root=''):
        self.UPGRADE_file = '/cm/UPGRADE'
        self.root_bashrc = '/root/.bashrc'
        self.my_cnf = '/etc/my.cnf'
        self.motd_file = '/etc/motd'
        self.cmid_file = '/cm/CLUSTERMANAGERID'

        if any(os.path.isfile(s) for s in ['/etc/SuSE-release', '/etc/SUSE-brand']):
            self.repo_dir = '/etc/zypp/repos.d'
            self.repo_services_dir = '/etc/zypp/services.d'
            self.repo_credentials_dir = '/etc/zypp/credentials.d'
            self.dvd_repo_file = self.repo_dir + '/cm-%s-dvd.repo' % (self.upgrade_version)
            self.prereq_dist_repo_file = self.repo_dir + '/prereq-dist.repo'
            self.main_cmd = 'zypper -n --no-gpg-checks'
            self.remove_cmd = (self.main_cmd + ' remove').split(' ')
            self.upgrade_cmd = (self.main_cmd + ' update').split(' ')
            self.downgrade_cmd = (self.main_cmd + ' install --oldpackage').split(' ')
            self.update_cmd = (self.main_cmd + ' update').split(' ')
            self.install_cmd = (self.main_cmd + ' install').split(' ')
            self.reinstall_cmd = (self.main_cmd + ' install -f').split(' ')
            self.repo_refresh = (self.main_cmd + ' refresh -f').split(' ')
            self.repo_clean = (self.main_cmd + ' clean --all').split(' ')
            self.repo_lock = (self.main_cmd + ' addlock').split(' ')
            self.repo_unlock = (self.main_cmd + ' removelock').split(' ')
            self.list_updates = (self.main_cmd + ' search -u -s --match-exact').split(' ')
            self.list_available = (self.main_cmd + ' search -u -s --match-exact').split(' ')
            self.current_cm_repos = ['%s/Cluster_Manager_Updates.repo' % (self.repo_dir),
                                     '%s/Cluster_Manager_NI_Updates.repo' % (self.repo_dir),
                                     '%s/Cluster_Manager_ML_Updates.repo' % (self.repo_dir)]
            self.disable_repo_cmd = (self.main_cmd + ' modifyrepo -d').split(' ')
            self.install_root_cmd = (self.main_cmd + ' --root').split(' ')
            self.install_root_cmd_rpm = ('rpm --root').split(' ')
            self.download_cache = '/root/pkgs-cache'
            self.download_pkg = (self.main_cmd + ' --pkg-cache-dir {} download'.format(self.download_cache)).split(' ')
            self.package_erase = ('rpm --erase').split(' ')
        elif self.is_ubuntu():
            self.repo_dir = '/etc/apt/sources.list.d'
            self.dvd_repo_file = self.repo_dir + '/cm-%s-dvd.list' % (self.upgrade_version)
            self.main_cmd = ['apt-get',
                             '-o', 'Dpkg::Options::=--force-confold',
                             '--allow-unauthenticated',
                             '-y']
            self.upgrade_cmd = self.main_cmd + ['install']
            self.update_cmd = self.main_cmd + ['install']
            self.install_cmd = self.main_cmd + ['install']
            self.repo_clean = 'apt update'.split(' ')
            self.repo_refresh = 'apt update'.split(' ')
            self.current_cm_repos = ['%s/cm.list' % (self.repo_dir),
                                     '%s/cm-ml.list' % (self.repo_dir)]
            self.repo_lock = 'apt-mark hold'.split(' ')
            self.list_updates = None
            self.cmconf_motd_file = '/cm/conf/etc/motd.%s' % (self.base_dist)
            self.motd_file = '/etc/update-motd.d/99-bright-cmd'
            self.prereq_dist_repo_file = self.repo_dir + '/prereq-dist.list'
            self.package_erase_nodeps = 'dpkg -r --force-depends'.split(' ')
            self.fix_broken_install = self.main_cmd + ['--fix-broken', 'install']
            self.list_available = ['apt-cache', 'search', '--names-only']
        else:
            self.repo_dir = '/etc/yum.repos.d'
            self.dvd_repo_file = self.repo_dir + '/cm-%s-dvd.repo' % (self.upgrade_version)
            self.prereq_dist_repo_file = self.repo_dir + '/prereq-dist.repo'
            self.main_cmd = 'yum -y --nogpgcheck --exclude=mariadb*'
            self.install_root_cmd = (self.main_cmd + ' --installroot').split(' ')
            self.install_root_cmd_rpm = ('rpm --root').split(' ')
            if len(self.disable_repos_list) > 0:
                self.main_cmd = self.main_cmd + ' ' + ' '.join(['--disablerepo=%s' % (rep) for rep in self.disable_repos_list])
            self.upgrade_cmd = (self.main_cmd + ' upgrade').split(' ')
            self.downgrade_cmd = (self.main_cmd + ' downgrade').split(' ')
            self.update_cmd = (self.main_cmd + ' update').split(' ')
            self.install_cmd = (self.main_cmd + ' install').split(' ')
            self.reinstall_cmd = (self.main_cmd + ' reinstall').split(' ')
            self.remove_cmd = (self.main_cmd + ' remove').split(' ')
            self.repo_refresh = None
            self.repo_clean = 'yum clean all'.split(' ')
            self.list_updates = (self.main_cmd + ' list updates').split(' ')
            self.list_available = (self.main_cmd + ' list available').split(' ')
            self.current_cm_repos = ['%s/cm.repo' % (self.repo_dir),
                                     '%s/cm-ni.repo' % (self.repo_dir),
                                     '%s/cm-ml.repo' % (self.repo_dir)]
            self.disable_repo_cmd = None
            self.download_cache = '/root/pkgs-cache'
            self.download_pkg = ('yumdownloader --destdir={}'.format(self.download_cache)).split(' ')
            self.package_erase = 'rpm --erase'.split(' ')

    def read_package_info(self, yaml_file):
        try:

            import yaml
            if not os.path.isfile(os.path.abspath(yaml_file)):
                print('%s does not exist, cannot read package list' % (yaml_file))
                sys.exit(1)

            logger.Log('Using package config %s' % (yaml_file))
            try:
                with open(yaml_file, 'r') as _package_info:
                    return yaml.load(_package_info)
            except yaml.YAMLError as yaml_error:
                print(str(yaml_error))
                sys.exit(1)

        except Exception as e:
            print(str(e))
            sys.exit(1)

    def is_rhel7(self):
        res = False
        try:
            if (self.base_dist is not None) and ('rhel7' in self.base_dist or 'centos7' in self.base_dist or 'sl7' in self.base_dist):
                res = True
        except Exception as e:
            res = False
            log_and_append(str(e), False, self.errors)
        return res

    def is_rhel6(self):
        res = False
        try:
            if (self.base_dist is not None) and ('rhel6' in self.base_dist or 'centos6' in self.base_dist or 'sl6' in self.base_dist):
                res = True
        except Exception as e:
            res = False
            log_and_append(str(e), False, self.errors)
        return res

    def is_rhel8(self):
        if self.base_dist is not None:
            return ('rhel8' in self.base_dist) or ('centos8' in self.base_dist)
        return False

    def is_rhel(self):
        return self.is_rhel7() or self.is_rhel6() or self.is_rhel8()

    def is_sles(self):
        return ((self.base_dist is not None) and ('sles' in self.base_dist))

    def is_ubuntu16(self):
        return ((self.base_dist is not None) and ('ubuntu16' in self.base_dist))

    def is_ubuntu18(self):
        return ((self.base_dist is not None) and ('ubuntu18' in self.base_dist))

    def is_ubuntu(self):
        return self.is_ubuntu16() or self.is_ubuntu18()

    def is_sles12(self):
        return ((self.base_dist is not None) and ('sles12' in self.base_dist))

    def get_all_image_dirs(self, ilist=[]):
        result = []
        try:
            if len(ilist) == 0:
                if len(self.additional_image_base_dirs) > 0:
                    for d in self.additional_image_base_dirs:
                        for imgdir in os.listdir(d):
                            result.append(d + '/' + imgdir)
                else:
                    for imgdir in os.listdir(self.softwareimages_maindir):
                        result.append(self.softwareimages_maindir + '/' + imgdir)
            else:
                for img in ilist:
                    if len(self.additional_image_base_dirs) > 0:
                        for d in self.additional_image_base_dirs:
                            result.append(d + '/' + img)
                    else:
                        result.append(self.softwareimages_maindir + '/' + img)
        except Exception as e:
            print(str(e))
            log_and_append(str(e), False, self.errors)
        return result

    def bf(self, txt):
        return Font.BOLD + txt + Font.CLOSE

    def cf(self, txt, color):
        return Font.__dict__[color] + txt + Font.CLOSE

    def display_prerequisites(self):
        print(self.bf('\n*** IMPORTANT: Prerequisites for upgrade ***\n'))
        print("""\
The following prerequisites must be met before starting the upgrade:

* All compute nodes must be powered off
* All cloud nodes and cloud directors must be terminated
* Existing updates must be applied to the head node and software images
* Recommended: Create full back-up of head node(s) if possible
* Extra base distribution packages will be installed, please make sure that
  appropriate repositories/software channels are enabled
* HA configuration: If shared storage of /cm/shared is not NFS mounted before
  performing the upgrade, please make sure that the data from the shared storage
  is in sync with local storage, or alternatively keep the shared storage mounted
  and perform an in-place upgrade.
* Please refer the upgrade manual for detailed information.
  https://support.brightcomputing.com/upgrade-manuals/9.1
  """)

    def intro(self):
        print(self.bf('DESCRIPTION'))
        print("""\
This is the Bright Cluster Manager upgrade utility. This can be used to upgrade a
Bright cluster from one major version to another (e.g., 9.0 to 9.1). It must be run
on both the active and passive head nodes when a cluster is running in High
Availability mode. It can be executed simultaneously or at different times.  It is
distribution independent and hence can be executed on both RedHat and SuSE Linux based
 Bright clusters. Software images can only be updated when the script is being run from
 the active head node.
    """)

    def show_manpage(self):
        self.intro()
        self.display_prerequisites()
        print(self.bf('HEAD NODE UPGRADE'))
        print("""\
The following is an outline of the main actions that are performed
during upgrade of head nodes

* Create back-up of cluster configuration
* Create back-up workload manager configurations where possible
* Upgrade all Bright RPMs to the new version
* Remove obsoleted RPMs if any
* Install new Bright RPMs if required
* Install new base distribution RPMS if required
* Upgrade CUDA RPMs (if --upgrade-cuda is specified)
* Update CMDaemon configuration file
* Collect information about rpmnew files and other configuration file changes
        """)

        print(self.bf('SOFTWARE IMAGE UPGRADE'))
        print("""\
The following is an outline of the main actions that are performed
during upgrade of all software images

* Upgrade all Bright RPMs to the new version
* Remove obsoleted RPMs if any
* Install new Bright RPMs if required
* Install new base distribution RPMS if required
* Update CMDaemon configuration file
* Collect information about rpmnew files and other configuration file changes.
        """)

        print(self.bf('LOG FILES'))
        print("""\
Head node upgrade:

* /var/log/cm-upgrade.log

Software image upgrade:

* <image directory>/var/log/cm-upgrade.log
    where <image directory> is for example: /cm/images/default-image
    """)

        print(self.bf('LIMITATIONS'))
        print("""\
When using a Bright DVD/ISO to upgrade software images:

* It is not possible to update multiple images of different base distributions at
the same time. Only one ISO/DVD can be specified to the command line. Please see
section on %s and %s for more details.
    """ % (self.bf('USAGE'), self.bf('EXAMPLES')))

        print(self.bf('UNSUPPORTED LINUX DISTRIBUTIONS'))
        print("""\
* The following Linux base distributions are not supported:

    * RHEL5, SL5, CENTOS5
    * SLES11sp1, SLES11sp2
    * RHEL6u4 and older, SL6u4 and older, CENTOS6u4 and older
    """)

        print(self.bf('SUPPORT'))
        print("""\
The Bright Cluster Manager Administration Manual describes how to send bug reports
to Bright Computing. Please attach corresponding upgrade logfiles (see %s),
along with support requests related to upgrade.
    """ % (self.bf('LOG FILES')))


logger = Logger('/var/log/cm-upgrade.log')
